import ConstructorOf from '@thats/utils/types/ConstructorOf'
import uniq from 'lodash/uniq'
import { PerspectiveCamera } from 'three'
import KeyboardController from '../controllers/KeyboardController'
import Quest2Controller from '../controllers/Quest2Controller'
import RenderMode from '../enums/RenderMode'
import PlayerConsole from '../playertools/PlayerConsole'
import PlayerHand from '../playertools/PlayerHand'
import { XRUpdateParams } from './XRClass'
import XRController from './XRController'
import XRObject from './XRObject'
import XRProject from './XRProject'

export class XRPlayer extends XRObject {
	private xr: boolean
	private controller: XRController
	private project: XRProject
	camera: PerspectiveCamera
	debugConsole: PlayerConsole
	leftHand: PlayerHand
	rightHand: PlayerHand

	defaults() {
		return {
			...super.defaults(),
			xr: false,
		}
	}

	init() {
		super.init()

		// Controller
		this.setController(KeyboardController)

		// Camera
		this.camera = new PerspectiveCamera(50, 1, 0.01, 1000)
		this.add(this.camera)

		// Left Hand
		this.leftHand = new PlayerHand()
		this.camera.add(this.leftHand)
		this.rightHand = new PlayerHand()
		this.camera.add(this.rightHand)

		// Debug Console
		this.debugConsole = new PlayerConsole()
		this.debugConsole.position.set(0, 0, -2)
		this.debugConsole.log('Hello world')
		this.camera.add(this.debugConsole)

		// Set Size
		this.setSize(this.project.viewer?.dims.w || 100, this.project.viewer?.dims.h || 100)
	}

	update(o?: XRUpdateParams) {
		this.controller?.update(o)
	}

	setController(C: ConstructorOf<XRController>) {
		this.controller?.destroy()
		uniq(this.project.scene.controllerBehaviours)?.forEach(B => (C = B(C)))
		this.controller = new C({ project: this.project })
	}

	setRenderMode(v: RenderMode) {
		this.xr = v === RenderMode.XR
		if (this.camera) this.camera.position.y = this.xr ? 0 : 1
		this.setController(this.xr ? Quest2Controller : KeyboardController)
	}

	setSize(width: number, height: number) {
		this.camera.aspect = width / height
		this.camera.updateProjectionMatrix()
	}

	walkBy(v: number) {
		if (v) this.translateZ(v)
	}

	strafeBy(v: number) {
		if (v) this.translateX(v)
	}

	lookUpBy(v: number) {
		if (v) this.camera.rotateX(v)
	}

	yawBy(v: number) {
		if (v) this.rotateY(v)
	}

	pitchBy(v: number) {
		if (v) this.rotateX(v)
	}

	rollBy(v: number) {
		if (v) this.rotateZ(v)
	}

	destroy() {
		this.controller?.destroy()
		this.leftHand?.destroy()
		this.rightHand?.destroy()
		this.debugConsole?.destroy()
	}
}
export default XRPlayer
