import ConstructorOf from '@thats/utils/types/ConstructorOf'
import { Clock, XRSession } from 'three'
import GrabActionType from '../enums/GrabActionType'
import PlayState from '../enums/PlayState'
import RenderMode from '../enums/RenderMode'
import setVector from '../utils/setVector'
import XRClass, { XRUpdateParams } from './XRClass'
import XRParam from './XRParam'
import XRPlayer from './XRPlayer'
import XRScene from './XRScene'
import XRViewer from './XRViewer'

export interface XRProjectState {
	play: PlayState
	grabAction: GrabActionType
	godMode: boolean
}

export default class XRProject extends XRClass() {
	private _state: XRProjectState
	private Scene: ConstructorOf<XRScene>
	private clock: Clock
	scene: XRScene
	session: XRSession
	player: XRPlayer
	viewer: XRViewer

	get state() {
		return this._state
	}

	setPlayState(v: PlayState) {
		this._state.play = v
	}

	setGrabActionState(v: GrabActionType) {
		this._state.grabAction = v
	}

	setGodModeState(v: boolean) {
		this._state.godMode = v
	}

	setScene(S: ConstructorOf<XRScene>) {
		this.Scene = S
		this.start()
	}

	defaults() {
		return {
			...super.defaults(),
			clock: new Clock(),
			_state: {
				play: PlayState.STOPPED,
				grabAction: GrabActionType.MOVE,
				godMode: false,
			},
			viewers: [],
		}
	}

	init() {
		this._endXR = this.endXR.bind(this)
		this._tock = this.tock.bind(this)
		super.init()
	}

	initViewer(el: HTMLElement, dims: any) {
		this.viewer = new XRViewer({ project: this, el, dims })
		this.play()
	}

	setParam<T>(param: XRParam<T>, value: T) {
		this.scene.setParam(param, value)
	}

	start() {
		try {
			this.viewer?.destroy()
			this.scene?.destroy()
			this.player?.destroy()
		} catch (e) {}

		// Scene
		this.scene = new this.Scene()

		// Player
		this.player = new XRPlayer({ project: this })
		this.player.setRenderMode(RenderMode.DESKTOP)
		setVector(this.player.position, this.scene.playerInitPosition)
		setVector(this.player.rotation, this.scene.playerInitRotation)
		this.scene.addObject(this.player)

		// Viewer
		this.viewer?.init()

		// Play
		this.clock.start()
		this._state.play = PlayState.PLAYING
		this.untick()
		this.tick()
	}

	stop() {
		this.pause()
		this.clock.stop()
		this._state.play = PlayState.STOPPED
		this.scene.destroy()
	}

	pause() {
		this._state.play = PlayState.PAUSED
	}

	play() {
		this._state.play = PlayState.PLAYING
		this.tick()
	}

	startXR() {
		this.pause()
		navigator['xr']
			.requestSession('immersive-vr', {
				optionalFeatures: ['local-floor', 'bounded-floor', 'hand-tracking', 'layers'],
			})
			.then(s => {
				this.session = s
				this.player.setRenderMode(RenderMode.XR)
				s.addEventListener('end', this._endXR)
				return this.viewer.setXRSession(s)
			})
			.then(() => this.play())
	}

	_endXR: any
	endXR() {
		this.pause()
		this.session.removeEventListener('end', this._endXR)
		this.player.setRenderMode(RenderMode.DESKTOP)
		this.session = undefined
		void this.viewer.setXRSession()
		this.play()
	}

	destroy() {
		this.stop()
		this.session?.removeEventListener('end', this._endXR)
		this.viewer?.destroy()
		this.player?.destroy()
	}

	private tick() {
		if (this.session) {
			this.session.requestAnimationFrame(this._tock)
		} else {
			window.requestAnimationFrame(this._tock)
		}
	}

	private untick() {
		if (this.session) {
			this.session.cancelAnimationFrame(this._tock)
		} else {
			window.cancelAnimationFrame(this._tock)
		}
	}
	private _tock: any
	private tock(_, frame) {
		try {
			if (this._state.play === PlayState.STOPPED) return
			const p: XRUpdateParams = {
				delta: this.clock.getDelta(),
				time: this.clock.getElapsedTime(),
				frame,
			}

			if (this._state.play !== PlayState.PAUSED) this.scene?.update(p)
			this.player?.update(p)
			this.viewer?.update(p)
		} catch (e) {
			console.error('Error in XRProject.tock', e)
		}
		this.tick()
	}
}
