import debounce from 'lodash/debounce'
import { Quaternion, Scene as TScene, Vector3 } from 'three'
import MotionType from '../enums/MotionType'
import ParamType from '../enums/ParamType'
import ButtonParam from '../params/ButtonParam'
import applyPosQuatScaleMargin from '../utils/applyPosQuatScaleMargin'
import collisionShapeToRigidBody from '../utils/collisionShapeToRigidBody'
import createCollisionShape from '../utils/createCollisionShape'
import createPhysicsWorld from '../utils/physics/createPhysicsWorld'
import destroyPhysicsWorld from '../utils/physics/destroyPhysicsWorld'
import vector3T2A from '../utils/vector3T2A'
import XRClass, { XRUpdateParams } from './XRClass'
import Light from './XRLight'
import XRMass from './XRMass'
import XRObject from './XRObject'
import XRParam from './XRParam'
import XRPlayer from './XRPlayer'

export default class XRScene extends XRClass() {
	key: string
	label: string
	playerInitPosition: Vector3
	playerInitRotation: Vector3
	gravity: Vector3
	params: { [key: string]: XRParam<any> }
	controllerBehaviours: any[]
	objects: XRObject[]
	masses: XRMass[]
	lights: Light[]
	gfxScene: TScene
	physxScene: Ammo.btSoftRigidDynamicsWorld

	defaults() {
		return {
			...super.defaults(),
			label: 'XRScene',
			playerInitPosition: new Vector3(0, 0, 10),
			playerInitRotation: new Vector3(0, 0, 0),
			gravity: new Vector3(0, -9.81, 0),
			params: {},
			controllerBehaviours: [],
			objects: [],
			masses: [],
			lights: [],
		}
	}

	init() {
		super.init()
		this.physxScene = createPhysicsWorld(vector3T2A(this.gravity))
		Object.keys(this.params).forEach(k => (this.params[k].key = k))
		this.gfxScene = new TScene()
		this.build()
	}

	build() {}

	update(p?: XRUpdateParams) {
		super.update(p)
		// If time has moved on, adjust the physics
		if (p?.delta) this.physxScene.stepSimulation(p?.delta)
		void Promise.all([
			...this.objects.map(o => o?.update(p)),
			...this.masses.map(o => o?.update(p)),
			...this.lights.map(o => o?.update(p)),
		])
	}

	reinit = debounce(() => {
		this.build()
	}, 700)

	setParam<T>(param: XRParam<T>, value: T) {
		if (param.type === ParamType.BUTTON) return void (param as ButtonParam).callback(this)
		this.params[param.key].value = value
		this.destroy(false)
		this.reinit()
	}

	get all() {
		return [...this.objects, ...this.masses, ...this.lights]
	}

	addLight(l: Light) {
		this.lights.push(l)
		this.gfxScene.add(l)
	}

	addObject(o: XRObject) {
		this.objects.push(o)
		this.gfxScene.add(o)
	}

	addMass(o: XRMass, type: MotionType = MotionType.Dynamic): any {
		this.masses.push(o)
		this.gfxScene.add(o)

		const pos = o.getWorldPosition(new Vector3())
		const quat = o.getWorldQuaternion(new Quaternion())
		const scale = o.getWorldScale(new Vector3())
		const collisionGroup = 1
		const collisionMask = -1

		// const p = prepareThreeObjectForCollisionShape(o)
		const collisionShape = createCollisionShape(o)
		const localTransform = applyPosQuatScaleMargin(collisionShape, pos, quat, scale)
		const rigidBody = collisionShapeToRigidBody(
			collisionShape,
			localTransform,
			o.mass,
			type === MotionType.Kinematic
		)

		o.body = rigidBody
		this.physxScene.addRigidBody(rigidBody, collisionGroup, collisionMask)
	}

	remove(o: XRObject | XRMass | Light) {
		this.gfxScene.remove(o)
		if (o instanceof XRMass) {
			if (o?.body) {
				this.physxScene.removeRigidBody(o.body)
				Ammo.destroy(o.body)
				delete o.body
			}
			this.masses = this.masses.filter(m => m !== o)
		} else if (o instanceof XRObject) {
			this.objects = this.objects.filter(m => m !== o)
		} else {
			this.lights = this.lights.filter(l => l !== o)
		}
	}

	destroy(fullyDestroy: boolean = true) {
		const player = this.objects.find(v => v instanceof XRPlayer)

		this.objects.filter(v => fullyDestroy || !(v === player)).forEach(this.remove.bind(this))
		this.masses.forEach(this.remove.bind(this))
		this.lights.forEach(this.remove.bind(this))
		if (fullyDestroy) destroyPhysicsWorld(this.physxScene)
		this.objects = []
		this.masses = []
		this.lights = []

		if (!fullyDestroy) this.objects.push(player)
		if (fullyDestroy) this.gfxScene = undefined
		if (fullyDestroy) this.controllerBehaviours = []
		if (fullyDestroy) this.params = {}
	}
}
