"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const three_1 = require("three");
const EffectComposer_1 = require("three/examples/jsm/postprocessing/EffectComposer");
const RenderPass_1 = require("three/examples/jsm/postprocessing/RenderPass");
const ShaderPass_1 = require("three/examples/jsm/postprocessing/ShaderPass");
const UnrealBloomPass_1 = require("three/examples/jsm/postprocessing/UnrealBloomPass");
const KeyCode_1 = __importDefault(require("../enums/KeyCode"));
const Layer_1 = require("../enums/Layer");
const XRClass_1 = __importDefault(require("./XRClass"));
class XRViewer extends (0, XRClass_1.default)() {
    setFullscreen(on) {
        this.fullscreen = on;
        this.onChangeFullscreen();
    }
    setDims(d) {
        this.dims = d;
        this.onChangeDims();
    }
    defaults() {
        this.renderer = new three_1.WebGLRenderer({ antialias: false, alpha: false });
        return Object.assign(Object.assign({}, super.defaults()), { enabled: true, shadows: true, fullscreen: false, bloom: {
                enabled: true,
                strength: 1,
                threshold: 0,
                radius: 1,
            } });
    }
    init() {
        this._handleKeyDown = this.handleKeyDown.bind(this);
        window.addEventListener('keydown', this._handleKeyDown);
        super.init();
        this.renderer.outputEncoding = three_1.sRGBEncoding;
        if (this.shadows) {
            this.renderer.shadowMap.enabled = true;
        }
        this.canvas = this.renderer.domElement;
        this.el.innerHTML = '';
        this.el.appendChild(this.canvas);
        this.bloomLayer = new three_1.Layers();
        this.bloomLayer.set(Layer_1.Layer.BLOOM);
        this.bloomComposer = new EffectComposer_1.EffectComposer(this.renderer);
        const renderPass = new RenderPass_1.RenderPass(this.project.scene.gfxScene, this.project.player.camera);
        this.bloomComposer.renderToScreen = false;
        this.bloomComposer.addPass(renderPass);
        if (this.bloom.enabled) {
            this.bloomComposer.addPass(new UnrealBloomPass_1.UnrealBloomPass(new three_1.Vector2(this.dims.w, this.dims.h), this.bloom.strength, this.bloom.radius, this.bloom.threshold));
        }
        const shaderPass = new ShaderPass_1.ShaderPass(new three_1.ShaderMaterial({
            uniforms: {
                resolution: { value: { x: 0.001, y: 0.001 } },
                baseTexture: { value: null },
                bloomTexture: { value: this.bloomComposer.renderTarget2.texture },
            },
            vertexShader: XRViewer.vertexShader,
            fragmentShader: XRViewer.fragmentShader,
            defines: {},
        }), 'baseTexture');
        shaderPass.needsSwap = true;
        this.finalComposer = new EffectComposer_1.EffectComposer(this.renderer);
        this.finalComposer.addPass(renderPass);
        this.finalComposer.addPass(shaderPass);
        this.onChangeFullscreen();
        this.onChangeDims();
    }
    update(p) {
        this.renderBloom();
        this.finalComposer.render();
    }
    destroy() {
        window.removeEventListener('keydown', this._handleKeyDown);
        this.el.innerHTML = '';
        this.canvas = null;
        this.bloomComposer.passes = [];
        this.finalComposer.passes = [];
    }
    setXRSession(s) {
        this.renderer.xr.enabled = !!s;
        return this.renderer.xr.setSession(s);
    }
    renderBloom() {
        this.project.scene.gfxScene.traverse(o => {
            if (!this.bloomLayer.test(o.layers)) {
                if (o['material'])
                    o['material'].visible = false;
            }
        });
        const oldBG = this.project.scene.gfxScene.background;
        const oldFog = this.project.scene.gfxScene.fog;
        this.project.scene.gfxScene.background = new three_1.Color(0x000000);
        this.project.scene.gfxScene.fog = undefined;
        this.bloomComposer.render();
        this.project.scene.gfxScene.background = oldBG;
        this.project.scene.gfxScene.fog = oldFog;
        this.project.scene.gfxScene.traverse(o => {
            if (!this.bloomLayer.test(o.layers)) {
                if (o['material'])
                    o['material'].visible = true;
            }
        });
    }
    handleKeyDown(e) {
        if (e.keyCode === KeyCode_1.default.ESC)
            this.onEscapePressed();
    }
    onEscapePressed() {
        if (this.fullscreen)
            this.fullscreen = false;
    }
    onChangeFullscreen() {
        if (this.fullscreen) {
            this.preDims = this.dims;
            Object.assign(this.canvas.style, {
                position: 'absolute',
                top: 0,
                left: 0,
                right: 0,
                bottom: 0,
                zIndex: 2000,
            });
            this.el.removeChild(this.canvas);
            this.dims = { w: window.innerWidth, h: window.innerHeight, x: 0, y: 0 };
            document.body.appendChild(this.canvas);
        }
        else if (this.canvas.parentElement === document.body) {
            this.canvas.style.position = 'initial';
            document.body.removeChild(this.canvas);
            this.dims = this.preDims;
            this.el.appendChild(this.canvas);
        }
    }
    onChangeDims() {
        var _a, _b;
        const { w: width, h: height } = this.dims;
        this.renderer.setPixelRatio(width / height);
        this.renderer.setSize(width, height);
        this.bloomComposer.setSize(width, height);
        this.finalComposer.setSize(width, height);
        (_b = (_a = this.project) === null || _a === void 0 ? void 0 : _a.player) === null || _b === void 0 ? void 0 : _b.setSize(width, height);
    }
}
exports.default = XRViewer;
XRViewer.vertexShader = `
			varying vec2 vUv;
			void main() {
				vUv = uv;
				gl_Position = projectionMatrix * modelViewMatrix * vec4( position, 1.0 );
			}`;
XRViewer.fragmentShader = `
			uniform sampler2D baseTexture;
			uniform sampler2D bloomTexture;
			varying vec2 vUv;
			void main() {
				gl_FragColor = ( texture2D( baseTexture, vUv ) + vec4( 1.0 ) * texture2D( bloomTexture, vUv ) );
			}`;
//# sourceMappingURL=XRViewer.js.map