import XYWH from '@thats/utils/types/XYWH'
import {
	Color,
	Layers,
	ShaderMaterial,
	sRGBEncoding,
	Vector2,
	WebGLRenderer,
	XRSession,
} from 'three'
import { EffectComposer } from 'three/examples/jsm/postprocessing/EffectComposer'
import { RenderPass } from 'three/examples/jsm/postprocessing/RenderPass'
import { ShaderPass } from 'three/examples/jsm/postprocessing/ShaderPass'
import { UnrealBloomPass } from 'three/examples/jsm/postprocessing/UnrealBloomPass'
import CameraAngle from '../enums/CameraAngle'
import KeyCode from '../enums/KeyCode'
import { Layer } from '../enums/Layer'
import XRClass, { XRUpdateParams } from './XRClass'
import XRProject from './XRProject'

interface BloomSettings {
	enabled: boolean
	strength: number
	threshold: number
	radius: number
}

export default class XRViewer extends XRClass() {
	private preDims: XYWH
	private project: XRProject
	private renderer: WebGLRenderer
	private bloomComposer: EffectComposer
	private finalComposer: EffectComposer
	private bloomLayer: Layers
	private fullscreen: boolean

	static vertexShader = `
			varying vec2 vUv;
			void main() {
				vUv = uv;
				gl_Position = projectionMatrix * modelViewMatrix * vec4( position, 1.0 );
			}`

	static fragmentShader = `
			uniform sampler2D baseTexture;
			uniform sampler2D bloomTexture;
			varying vec2 vUv;
			void main() {
				gl_FragColor = ( texture2D( baseTexture, vUv ) + vec4( 1.0 ) * texture2D( bloomTexture, vUv ) );
			}`

	dims: XYWH
	el: HTMLElement
	canvas: HTMLCanvasElement
	shadows: boolean
	cameraAngle: CameraAngle
	enabled: boolean
	bloom: BloomSettings

	setFullscreen(on: boolean) {
		this.fullscreen = on
		this.onChangeFullscreen()
	}

	setDims(d: XYWH) {
		this.dims = d
		this.onChangeDims()
	}

	defaults() {
		this.renderer = new WebGLRenderer({ antialias: false, alpha: false })
		return {
			...super.defaults(),
			enabled: true,
			shadows: true,
			fullscreen: false,
			bloom: {
				enabled: true,
				strength: 1,
				threshold: 0,
				radius: 1,
			},
		}
	}

	init() {
		this._handleKeyDown = this.handleKeyDown.bind(this)
		window.addEventListener('keydown', this._handleKeyDown)

		super.init()

		// Renderer
		this.renderer.outputEncoding = sRGBEncoding
		// this.renderer.toneMapping = ReinhardToneMapping
		if (this.shadows) {
			this.renderer.shadowMap.enabled = true
			// this.renderer.shadowMap.type = BasicShadowMap
		}

		// Canvas
		this.canvas = this.renderer.domElement

		// Dom
		this.el.innerHTML = ''
		this.el.appendChild(this.canvas)

		// Layers
		this.bloomLayer = new Layers()
		this.bloomLayer.set(Layer.BLOOM)

		// Post Processing Composer
		this.bloomComposer = new EffectComposer(this.renderer)
		const renderPass = new RenderPass(this.project.scene.gfxScene, this.project.player.camera)
		this.bloomComposer.renderToScreen = false
		this.bloomComposer.addPass(renderPass)

		if (this.bloom.enabled) {
			this.bloomComposer.addPass(
				new UnrealBloomPass(
					new Vector2(this.dims.w, this.dims.h),
					this.bloom.strength,
					this.bloom.radius,
					this.bloom.threshold
				)
			)
		}

		const shaderPass = new ShaderPass(
			new ShaderMaterial({
				uniforms: {
					resolution: { value: { x: 0.001, y: 0.001 } },
					baseTexture: { value: null },
					bloomTexture: { value: this.bloomComposer.renderTarget2.texture },
				},
				vertexShader: XRViewer.vertexShader,
				fragmentShader: XRViewer.fragmentShader,
				defines: {},
			}),
			'baseTexture'
		)
		shaderPass.needsSwap = true

		this.finalComposer = new EffectComposer(this.renderer)
		this.finalComposer.addPass(renderPass)
		this.finalComposer.addPass(shaderPass)
		this.onChangeFullscreen()
		this.onChangeDims()
	}

	update(p?: XRUpdateParams) {
		this.renderBloom()
		this.finalComposer.render()
	}

	destroy() {
		window.removeEventListener('keydown', this._handleKeyDown)
		this.el.innerHTML = ''
		this.canvas = null
		this.bloomComposer.passes = []
		this.finalComposer.passes = []
	}

	setXRSession(s?: XRSession) {
		this.renderer.xr.enabled = !!s
		return this.renderer.xr.setSession(s)
	}

	private renderBloom() {
		this.project.scene.gfxScene.traverse(o => {
			if (!this.bloomLayer.test(o.layers)) {
				if (o['material']) o['material'].visible = false
			}
		})
		const oldBG = this.project.scene.gfxScene.background
		const oldFog = this.project.scene.gfxScene.fog
		this.project.scene.gfxScene.background = new Color(0x000000)
		this.project.scene.gfxScene.fog = undefined
		this.bloomComposer.render()
		this.project.scene.gfxScene.background = oldBG
		this.project.scene.gfxScene.fog = oldFog
		this.project.scene.gfxScene.traverse(o => {
			if (!this.bloomLayer.test(o.layers)) {
				if (o['material']) o['material'].visible = true
			}
		})
	}

	_handleKeyDown: any
	private handleKeyDown(e: KeyboardEvent) {
		if (e.keyCode === KeyCode.ESC) this.onEscapePressed()
	}

	private onEscapePressed() {
		if (this.fullscreen) this.fullscreen = false
	}

	private onChangeFullscreen() {
		if (this.fullscreen) {
			this.preDims = this.dims
			Object.assign(this.canvas.style, {
				position: 'absolute',
				top: 0,
				left: 0,
				right: 0,
				bottom: 0,
				zIndex: 2000,
			})
			this.el.removeChild(this.canvas)
			this.dims = { w: window.innerWidth, h: window.innerHeight, x: 0, y: 0 }
			document.body.appendChild(this.canvas)
		} else if (this.canvas.parentElement === document.body) {
			this.canvas.style.position = 'initial'
			document.body.removeChild(this.canvas)
			this.dims = this.preDims
			this.el.appendChild(this.canvas)
		}
	}

	private onChangeDims() {
		const { w: width, h: height } = this.dims
		this.renderer.setPixelRatio(width / height)
		this.renderer.setSize(width, height)
		this.bloomComposer.setSize(width, height)
		this.finalComposer.setSize(width, height)
		this.project?.player?.setSize(width, height)
	}
}
