import { Color, LinearFilter, Texture, Vector2 } from 'three'
import Plane from '../objects/Plane'

export default class PlayerConsole extends Plane {
	canvas: HTMLCanvasElement
	ctx: CanvasRenderingContext2D
	fontSizePx: number
	fontSizeWorld: number
	paneSize: Vector2
	textColor: number
	message: string = ''
	texture: Texture

	defaults() {
		return {
			...super.defaults(),
			paneSize: new Vector2(2, 2),
			fontSizePx: 50,
			fontSizeWorld: 0.15,
			textColor: 0x444444,
			visible: false,
		}
	}

	init() {
		super.init()
		this.canvas = document.createElement('canvas')
		this.ctx = this.canvas.getContext('2d')
		this.texture = new Texture(this.canvas)
		this.texture.minFilter = LinearFilter

		const worldToPx = this.fontSizeWorld / this.fontSizePx
		this.canvas.width = this.paneSize.x / worldToPx
		this.canvas.height = this.paneSize.y / worldToPx
		this.ctx.textAlign = 'center'
		this.ctx.textBaseline = 'middle'
		this.ctx.font = `${this.fontSizePx}px sans-serif`
		this.clearScreen()
	}

	show() {
		this.visible = true
	}

	hide() {
		this.visible = false
	}

	logVector3(v: { x: number; y: number; z: number }) {
		this.log(
			[Math.round(v.x * 10) / 10, Math.round(v.y * 10) / 10, Math.round(v.z * 10) / 10].join(
				', '
			)
		)
	}

	log(message: string) {
		const worldToPx = this.fontSizeWorld / this.fontSizePx
		this.clearScreen()

		const words = message.split(' ')
		const maxCols = this.paneSize.x / (this.fontSizeWorld * 0.6)
		const maxRows = this.paneSize.y / this.fontSizeWorld

		const rows = []
		let currentRow = ''
		words.forEach(word => {
			if (currentRow && currentRow.length + word.length > maxCols) {
				rows.push(currentRow)
				currentRow = word
			} else {
				currentRow += ' ' + word
			}
		})
		rows.push(currentRow)

		this.ctx.fillStyle = '#' + new Color(this.textColor).getHexString()
		rows.forEach((row, idx) => {
			this.ctx.fillText(
				row.trim(),
				this.paneSize.x / 2 / worldToPx,
				(idx + 1) * this.fontSizePx
			)
		})

		this.texture.needsUpdate = true

		this.material.map = this.texture
	}

	clearScreen() {
		this.ctx.fillStyle = '#' + this.material.color.getHexString()
		this.ctx.fillRect(0, 0, this.canvas.width, this.canvas.height)
	}
}
