import { MeshBasicMaterial, MeshLambertMaterial, Vector3 } from 'three'
import Walker from '../behaviours/walker/Walker'
import { XRUpdateParams } from '../core/XRClass'
import XRObject from '../core/XRObject'
import XRScene from '../core/XRScene'
import Box from '../objects/Box'
import PointLightBulb from '../objects/PointLightBulb'
import BooleanParam from '../params/BooleanParam'
import NumberParam from '../params/NumberParam'
import randomVector3 from '../utils/randomVector3'
import rotateBy from '../utils/rotateBy'
import setVector from '../utils/setVector'

export class Spinning extends XRScene {
	defaults() {
		return {
			...super.defaults(),
			key: 'spinning',
			label: 'Spinning',
			controllerBehaviours: Walker(true),
			params: {
				bodyDistance: new NumberParam({
					label: 'Body Distance',
					defaultValue: 4,
					min: 1,
					max: 20,
					step: 1,
				}),
				noOfLightBulbs: new NumberParam({
					label: 'No of Lightbulbs',
					defaultValue: 4,
					min: 1,
					max: 20,
					step: 1,
				}),
				noOfBoxes: new NumberParam({
					label: 'No of Boxes',
					defaultValue: 30,
					min: 1,
					max: 500,
					step: 10,
				}),
				boxSpeed: new NumberParam({
					label: 'Box Speed',
					defaultValue: 0.5, // % of PI per second
					min: 0.1,
					max: 1,
					step: 0.1,
				}),
				lightSpeed: new NumberParam({
					label: 'Light Speed',
					defaultValue: 0.5, // % of PI per second
					min: 0.1,
					max: 1,
					step: 0.1,
				}),
				boxColorRandom: new BooleanParam({ label: 'Random Box Color', defaultValue: true }),
				lightColorRandom: new BooleanParam({
					label: 'Random Light Color',
					defaultValue: true,
				}),
				boxSize: new NumberParam({
					label: 'Box Size',
					defaultValue: 1,
					min: 0.1,
					max: 10,
					step: 0.1,
				}),
				lightDistance: new NumberParam({
					label: 'Light Distance',
					defaultValue: 5,
					min: 1,
					max: 10,
					step: 1,
				}),
				lightDecay: new NumberParam({
					label: 'Light Decay',
					defaultValue: 1,
					min: 1,
					max: 4,
					step: 1,
				}),
			},
		}
	}

	init() {
		super.init()

		for (let i = 0; i < this.params.noOfLightBulbs.value; i++) {
			const color = this.params.lightColorRandom.value ? Math.random() * 0xffffff : 0xffffff
			const position = randomVector3(this.params.bodyDistance.value)
			const pointLight = new PointLightBulb({
				mass: 0,
				distance: this.params.lightDistance.value,
				decay: this.params.lightDecay.value,
				material: new MeshBasicMaterial({ color }),
			})
			const pointLightContainer = new XRObject({
				userData: {
					speed: randomVector3(this.params.lightSpeed.value),
				},
			})
			pointLightContainer.add(pointLight)
			setVector(pointLight.position, position)
			this.addObject(pointLightContainer)
		}

		for (let i = 0; i < this.params.noOfBoxes.value; i++) {
			const position = randomVector3(this.params.bodyDistance.value)
			const size = randomVector3(this.params.boxSize.value)
			const box = new Box({
				mass: 0,
				width: size.x,
				height: size.y,
				depth: size.z,
				userData: {
					speed: randomVector3(this.params.boxSpeed.value),
				},
				material: new MeshLambertMaterial({
					color: this.params.boxColorRandom.value ? Math.random() * 0xffffff : 0xffffff,
				}),
			})
			setVector(box.position, position)
			this.addObject(box)
		}
	}

	update(p?: XRUpdateParams) {
		super.update(p)

		if (p?.delta) {
			this.objects.forEach(obj => {
				if (obj.userData.speed) {
					rotateBy(
						obj,
						(obj.userData.speed as Vector3).clone().multiplyScalar(p.delta * Math.PI)
					)
				}
			})
		}
	}
}
export default Spinning
