import { BufferGeometry, Matrix4, Quaternion } from 'three'
import XRMass from '../core/XRMass'
import Box from '../objects/Box'
import Cone from '../objects/Cone'
import Cylinder from '../objects/Cylinder'
import Plane from '../objects/Plane'
import Sphere from '../objects/Sphere'
import Torus from '../objects/Torus'
import iterateGeometries from './iterateGeometries'
import createTorusShape from './physics/createTorusShape'
import createTriMeshShape from './physics/createTriMeshShape'

export default function createCollisionShape(o: XRMass): Ammo.btCollisionShape {
	const quat = o.quaternion || new Quaternion(0, 0, 0, 1)

	const geometry = o['geometry']
	if (o && geometry?.isGeometry) {
		o.geometry = new BufferGeometry()['fromGeometry'](geometry)
	}

	// prepare data to pass to three-to-ammo.js
	const extractData = (o: XRMass) => {
		const matrixWorld = new Matrix4().elements
		const vertices: any[] = []
		const matrices: any[] = []
		const indexes: any[] = []
		iterateGeometries(o, {}, (vertexArray: any, matrixArray: any, indexArray: any) => {
			vertices.push(vertexArray)
			matrices.push(matrixArray)
			indexes.push(indexArray)
		})

		return { vertices, matrices, indexes, matrixWorld }
	}

	const d = extractData(o)

	let collisionShape
	const btHalfExtents = new Ammo.btVector3()
	if (o instanceof Box) {
		btHalfExtents.setValue(o.width / 2, o.height / 2, o.depth / 2)
		collisionShape = new Ammo.btBoxShape(btHalfExtents)
	} else if (o instanceof Sphere) {
		collisionShape = new Ammo.btSphereShape(o.radius)
	} else if (o instanceof Cylinder) {
		btHalfExtents.setValue(o.radiusTop, o.height / 2, o.radiusBottom)
		collisionShape = new Ammo.btCylinderShape(btHalfExtents)
	} else if (o instanceof Cone) {
		collisionShape = new Ammo.btConeShape(o.radius, o.height)
		// } else if (o instanceof Capsule) {
		// 	collisionShape = new Ammo.btCapsuleShape(o.radius, o.height)
	} else if (o instanceof Torus) {
		collisionShape = createTorusShape(o, quat)
	} else if (o instanceof Plane) {
		collisionShape = createTriMeshShape(d.vertices, d.matrices, d.indexes, d.matrixWorld, {
			...o,
			concave: false,
		})
	}
	// collisionShape = createHullShape(d.vertices, d.matrices, d.matrixWorld, params)
	// collisionShape = createHACDShapes(
	// 	d.vertices,
	// 	d.matrices,
	// 	d.indexes,
	// 	d.matrixWorld,
	// 	params
	// )
	// collisionShape = createVHACDShapes(
	// 	d.vertices,
	// 	d.matrices,
	// 	d.indexes,
	// 	d.matrixWorld,
	// 	params
	// )
	//collisionShape = createTriMeshShape(d.vertices, d.matrices, d.indexes, d.matrixWorld, {
	// 				...params,
	// 				concave: false,
	// 			})
	//collisionShape = createTriMeshShape(d.vertices, d.matrices, d.indexes, d.matrixWorld, {
	// 				...params,
	// 				concave: true,
	// 			})

	// if there is a x, y or z, take is as temporary offset parameter
	// const { x, y, z } = params
	// if (x || y || z) {
	// 	collisionShape._compoundOffset = { x: x || 0, y: y || 0, z: z || 0 }
	// }

	// in some cases, like hacd, it will be an array of shapes
	// so we merge them
	// if (Array.isArray(collisionShape))
	// 	collisionShape = mergeCollisionShapesToCompoundShape(collisionShape)

	return collisionShape as Ammo.btCollisionShape
}
