import { Matrix4, Vector3 } from 'three'
import ShapeFit from '../../enums/ShapeFit'
import ShapeType from '../../enums/ShapeType'
import computeBounds from '../computeBounds'
import computeScale from '../computeScale'
import finishCollisionShape from '../finishCollisionShape'
import setDefaultOptions from './setDefaultOptions'

export const createHACDShapes = (function () {
	const vector = new Vector3()
	const center = new Vector3()
	const matrix = new Matrix4()
	return function (vertices, matrices, indexes, matrixWorld, options: any = {}) {
		options.type = ShapeType.HACD
		setDefaultOptions(options)

		if (options.fit === ShapeFit.MANUAL) {
			console.warn('cannot use fit: manual with type: hacd')
			return []
		}

		if (!Ammo.hasOwnProperty('HACD')) {
			console.warn(
				'HACD unavailable in included build of Ammo.js. Visit https://github.com/mozillareality/ammo.js for the latest version.'
			)
			return []
		}

		const bounds = computeBounds(vertices, matrices)
		const scale = computeScale(matrixWorld, options)

		let vertexCount = 0
		let triCount = 0
		center.addVectors(bounds.max, bounds.min).multiplyScalar(0.5)

		for (let i = 0; i < vertices.length; i++) {
			vertexCount += vertices[i].length / 3
			if (indexes && indexes[i]) {
				triCount += indexes[i].length / 3
			} else {
				triCount += vertices[i].length / 9
			}
		}

		const hacd = new Ammo['HACD']()
		if (options.hasOwnProperty('compacityWeight'))
			hacd.SetCompacityWeight(options.compacityWeight)
		if (options.hasOwnProperty('volumeWeight')) hacd.SetVolumeWeight(options.volumeWeight)
		if (options.hasOwnProperty('nClusters')) hacd.SetNClusters(options.nClusters)
		if (options.hasOwnProperty('nVerticesPerCH')) hacd.SetNVerticesPerCH(options.nVerticesPerCH)
		if (options.hasOwnProperty('concavity')) hacd.SetConcavity(options.concavity)

		const points = Ammo._malloc(vertexCount * 3 * 8)
		const triangles = Ammo._malloc(triCount * 3 * 4)
		hacd.SetPoints(points)
		hacd.SetTriangles(triangles)
		hacd.SetNPoints(vertexCount)
		hacd.SetNTriangles(triCount)

		let pptr = points / 8,
			tptr = triangles / 4

		for (let i = 0; i < vertices.length; i++) {
			const components = vertices[i]
			matrix.fromArray(matrices[i])
			for (let j = 0; j < components.length; j += 3) {
				vector
					.set(components[j + 0], components[j + 1], components[j + 2])
					.applyMatrix4(matrix)
					.sub(center)
				Ammo.HEAPF64[pptr + 0] = vector.x
				Ammo.HEAPF64[pptr + 1] = vector.y
				Ammo.HEAPF64[pptr + 2] = vector.z
				pptr += 3
			}
			if (indexes[i]) {
				const indices = indexes[i]
				for (let j = 0; j < indices.length; j++) {
					Ammo.HEAP32[tptr] = indices[j]
					tptr++
				}
			} else {
				for (let j = 0; j < components.length / 3; j++) {
					Ammo.HEAP32[tptr] = j
					tptr++
				}
			}
		}

		hacd.Compute()
		Ammo._free(points)
		Ammo._free(triangles)
		const nClusters = hacd.GetNClusters()

		const shapes = []
		for (let i = 0; i < nClusters; i++) {
			const hull = new Ammo.btConvexHullShape()
			hull.setMargin(options.margin)
			const nPoints = hacd.GetNPointsCH(i)
			const nTriangles = hacd.GetNTrianglesCH(i)
			const hullPoints = Ammo._malloc(nPoints * 3 * 8)
			const hullTriangles = Ammo._malloc(nTriangles * 3 * 4)
			hacd.GetCH(i, hullPoints, hullTriangles)

			const pptr = hullPoints / 8
			for (let pi = 0; pi < nPoints; pi++) {
				const btVertex = new Ammo.btVector3()
				const px = Ammo.HEAPF64[pptr + pi * 3 + 0]
				const py = Ammo.HEAPF64[pptr + pi * 3 + 1]
				const pz = Ammo.HEAPF64[pptr + pi * 3 + 2]
				btVertex.setValue(px, py, pz)
				hull.addPoint(btVertex, pi === nPoints - 1)
				Ammo.destroy(btVertex)
			}

			finishCollisionShape(hull, options, scale)
			shapes.push(hull)
		}

		return shapes
	}
})()
export default createHACDShapes
