import { Matrix4, Vector3 } from 'three'
import ShapeFit from '../../enums/ShapeFit'
import ShapeType from '../../enums/ShapeType'
import computeBounds from '../computeBounds'
import computeScale from '../computeScale'
import finishCollisionShape from '../finishCollisionShape'
import setDefaultOptions from './setDefaultOptions'

export const createHullShape = (function () {
	const vertex = new Vector3()
	const center = new Vector3()
	const matrix = new Matrix4()
	return function (vertices, matrices, matrixWorld, options: any = {}) {
		options.type = ShapeType.HULL
		setDefaultOptions(options)

		if (options.fit === ShapeFit.MANUAL) {
			console.warn('cannot use fit: manual with type: hull')
			return null
		}

		const bounds = computeBounds(vertices, matrices)

		const btVertex = new Ammo.btVector3()
		const originalHull = new Ammo.btConvexHullShape()
		originalHull.setMargin(options.margin)
		center.addVectors(bounds.max, bounds.min).multiplyScalar(0.5)

		let vertexCount = 0
		for (let i = 0; i < vertices.length; i++) {
			vertexCount += vertices[i].length / 3
		}

		const maxVertices = options.hullMaxVertices || 100000
		// todo: might want to implement this in a deterministic way that doesn't do O(verts) calls to Math.random
		if (vertexCount > maxVertices) {
			console.warn(
				`too many vertices for hull shape; sampling ~${maxVertices} from ~${vertexCount} vertices`
			)
		}
		const p = Math.min(1, maxVertices / vertexCount)

		for (let i = 0; i < vertices.length; i++) {
			const components = vertices[i]
			matrix.fromArray(matrices[i])
			for (let j = 0; j < components.length; j += 3) {
				const isLastVertex = i === vertices.length - 1 && j === components.length - 3
				if (Math.random() <= p || isLastVertex) {
					// always include the last vertex
					vertex
						.set(components[j], components[j + 1], components[j + 2])
						.applyMatrix4(matrix)
						.sub(center)
					btVertex.setValue(vertex.x, vertex.y, vertex.z)
					originalHull.addPoint(btVertex, isLastVertex) // recalc AABB only on last geometry
				}
			}
		}

		let collisionShape = originalHull
		if (originalHull.getNumVertices() >= 100) {
			//Bullet documentation says don't use convexHulls with 100 verts or more
			const shapeHull = new Ammo.btShapeHull(originalHull)
			shapeHull.buildHull(options.margin)
			Ammo.destroy(originalHull)
			collisionShape = new Ammo.btConvexHullShape(
				Ammo['getPointer'](shapeHull.getVertexPointer()),
				shapeHull.numVertices()
			)
			Ammo.destroy(shapeHull) // btConvexHullShape makes a copy
		}

		Ammo.destroy(btVertex)

		finishCollisionShape(collisionShape, options, computeScale(matrixWorld, options))
		return collisionShape
	}
})()
export default createHullShape
