"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.createVHACDShapes = void 0;
const three_1 = require("three");
const ShapeFit_1 = __importDefault(require("../../enums/ShapeFit"));
const ShapeType_1 = __importDefault(require("../../enums/ShapeType"));
const computeBounds_1 = __importDefault(require("../computeBounds"));
const computeScale_1 = __importDefault(require("../computeScale"));
const finishCollisionShape_1 = __importDefault(require("../finishCollisionShape"));
const setDefaultOptions_1 = __importDefault(require("./setDefaultOptions"));
exports.createVHACDShapes = (function () {
    const vector = new three_1.Vector3();
    const center = new three_1.Vector3();
    const matrix = new three_1.Matrix4();
    return function (vertices, matrices, indexes, matrixWorld, options = {}) {
        options.type = ShapeType_1.default.VHACD;
        (0, setDefaultOptions_1.default)(options);
        if (options.fit === ShapeFit_1.default.MANUAL) {
            console.warn('cannot use fit: manual with type: vhacd');
            return [];
        }
        if (!Ammo.hasOwnProperty('VHACD')) {
            console.warn('VHACD unavailable in included build of Ammo.js. Visit https://github.com/mozillareality/ammo.js for the latest version.');
            return [];
        }
        const bounds = (0, computeBounds_1.default)(vertices, matrices);
        const scale = (0, computeScale_1.default)(matrixWorld, options);
        let vertexCount = 0;
        let triCount = 0;
        center.addVectors(bounds.max, bounds.min).multiplyScalar(0.5);
        for (let i = 0; i < vertices.length; i++) {
            vertexCount += vertices[i].length / 3;
            if (indexes && indexes[i]) {
                triCount += indexes[i].length / 3;
            }
            else {
                triCount += vertices[i].length / 9;
            }
        }
        const vhacd = new Ammo['VHACD']();
        const params = new Ammo['Parameters']();
        if (options.hasOwnProperty('resolution'))
            params.set_m_resolution(options.resolution);
        if (options.hasOwnProperty('depth'))
            params.set_m_depth(options.depth);
        if (options.hasOwnProperty('concavity'))
            params.set_m_concavity(options.concavity);
        if (options.hasOwnProperty('planeDownsampling'))
            params.set_m_planeDownsampling(options.planeDownsampling);
        if (options.hasOwnProperty('convexhullDownsampling'))
            params.set_m_convexhullDownsampling(options.convexhullDownsampling);
        if (options.hasOwnProperty('alpha'))
            params.set_m_alpha(options.alpha);
        if (options.hasOwnProperty('beta'))
            params.set_m_beta(options.beta);
        if (options.hasOwnProperty('gamma'))
            params.set_m_gamma(options.gamma);
        if (options.hasOwnProperty('pca'))
            params.set_m_pca(options.pca);
        if (options.hasOwnProperty('mode'))
            params.set_m_mode(options.mode);
        if (options.hasOwnProperty('maxNumVerticesPerCH'))
            params.set_m_maxNumVerticesPerCH(options.maxNumVerticesPerCH);
        if (options.hasOwnProperty('minVolumePerCH'))
            params.set_m_minVolumePerCH(options.minVolumePerCH);
        if (options.hasOwnProperty('convexhullApproximation'))
            params.set_m_convexhullApproximation(options.convexhullApproximation);
        if (options.hasOwnProperty('oclAcceleration'))
            params.set_m_oclAcceleration(options.oclAcceleration);
        const points = Ammo._malloc(vertexCount * 3 * 8 + 3);
        const triangles = Ammo._malloc(triCount * 3 * 4);
        let pptr = points / 8, tptr = triangles / 4;
        for (let i = 0; i < vertices.length; i++) {
            const components = vertices[i];
            matrix.fromArray(matrices[i]);
            for (let j = 0; j < components.length; j += 3) {
                vector
                    .set(components[j + 0], components[j + 1], components[j + 2])
                    .applyMatrix4(matrix)
                    .sub(center);
                Ammo.HEAPF64[pptr + 0] = vector.x;
                Ammo.HEAPF64[pptr + 1] = vector.y;
                Ammo.HEAPF64[pptr + 2] = vector.z;
                pptr += 3;
            }
            if (indexes[i]) {
                const indices = indexes[i];
                for (let j = 0; j < indices.length; j++) {
                    Ammo.HEAP32[tptr] = indices[j];
                    tptr++;
                }
            }
            else {
                for (let j = 0; j < components.length / 3; j++) {
                    Ammo.HEAP32[tptr] = j;
                    tptr++;
                }
            }
        }
        vhacd.Compute(points, 3, vertexCount, triangles, 3, triCount, params);
        Ammo._free(points);
        Ammo._free(triangles);
        const nHulls = vhacd.GetNConvexHulls();
        const shapes = [];
        const ch = new Ammo['ConvexHull']();
        for (let i = 0; i < nHulls; i++) {
            vhacd.GetConvexHull(i, ch);
            const nPoints = ch.get_m_nPoints();
            const hullPoints = ch.get_m_points();
            const hull = new Ammo.btConvexHullShape();
            hull.setMargin(options.margin);
            for (let pi = 0; pi < nPoints; pi++) {
                const btVertex = new Ammo.btVector3();
                const px = ch.get_m_points(pi * 3 + 0);
                const py = ch.get_m_points(pi * 3 + 1);
                const pz = ch.get_m_points(pi * 3 + 2);
                btVertex.setValue(px, py, pz);
                hull.addPoint(btVertex, pi === nPoints - 1);
                Ammo.destroy(btVertex);
            }
            (0, finishCollisionShape_1.default)(hull, options, scale);
            shapes.push(hull);
        }
        Ammo.destroy(ch);
        Ammo.destroy(vhacd);
        return shapes;
    };
})();
exports.default = exports.createVHACDShapes;
//# sourceMappingURL=createVHACDShapes.js.map